import {LayoutAnimation} from 'react-native';
import moment from "moment";
import {Linking} from "react-native";
import DeviceInfo from 'react-native-device-info';

import {Color, MomentFormat, JavascriptType, RecurrenceType, OpportunityState} from "./src/Enums";

export const ValueIsSet = (object) => {
    return object !== undefined && object !== null;
};

//This method allows for the copying of objects wih no connection by reference to the original object
//The downside to this method is that it only works on JSON safe objects (ex. if you copy an object that contains a moment, the copy will replace the moment with a date string)
//NOTE: If you are copying an array it is better to just use arrayName.slice()
export const DeepCopy = (object) => {
    return JSON.parse(JSON.stringify(object));
};

//This method allows for copying of objects, but any nested values will be copied by reference
//This copy is preferable to the DeepCopy as long as the nested values aren't going to cause an issue
//NOTE: If you are copying an array it is better to just use arrayName.slice()
export const ShallowCopy = (object) => {
    return Object.assign({}, object);
};

export const IsArrowFunction = (func) => {
    //If the type of func isn't "function" then it definitely isn't an arrow function
    if (typeof func !== JavascriptType.FUNCTION)
        return false;

    //Arrow functions don't have a prototype, whereas regular functions do
    //If the code is being transpiled into es5 this doesn't hold true
    if (!func.hasOwnProperty("prototype"))
        return true;

    //Because of automatic transpiling to es5 arrow functions are being converted to normal functions where all instances of 'this' are replaced with '_this'
    //Will want to revisit this eventually since this is messy
    //Note that if an arrow function doesn't contain any calls to 'this' in them we will return false, but that should be fine because an arrow function with no call to 'this' should be equivalent to a regular function anyway
    return func.toString().indexOf('_this') !== -1;
};

export const NumberToFixedDecimals = (number, decimals) => {
    return Number.parseFloat(number).toFixed(decimals);
};

export const GetNumberWithCommas = (number) => {
    return number.toString().replace(/\B(?=(\d{3})+(?!\d))/g, ",")
};

export const GetDollarString = (number, roundDown) => {
    if (number >= 0) {
        if (roundDown) {
            return '$' + GetNumberWithCommas(Math.floor(number));
        } else {
            return '$' + GetNumberWithCommas(NumberToFixedDecimals(number, 2));
        }
    } else {
        if (roundDown) {
            return '-$' + GetNumberWithCommas(Math.floor(number));
        } {
            return '-$' + GetNumberWithCommas(NumberToFixedDecimals(Math.abs(number), 2));
        }
    }
};

export const Uppercase = (string) => {
    return string.charAt(0).toUpperCase() + string.slice(1);
};

export const StringHasCharacters = (string) => {
    return string.trim() !== '';
};

export const Coalesce = (...objects) => {
    //This for loop structure is the most efficient way of iterating through an array in JavaScript
    for (let i = 0, n = objects.length; i < n; i++) {
        if (ValueIsSet(objects[i]))
            return objects[i];
    }

    return null;
};

export const StringHasCurrencyFormat = (string) => {
    return /^[+-]?[0-9]{1,3}(?:,?[0-9]{3})*(?:\.[0-9]{0,2})?$/.test(string);
};

export const StringHasNumberFormat = (string) => {
    return /^[+-]?[0-9]{1,3}(?:,?[0-9]{3})*(?:\.[0-9]*)?$/.test(string);
};

export const StringHasEmailFormat = (string) => {
    return /^[\w-\.]+@([\w-]+\.)+[\w-]{2,4}$/.test(string);
};

export const DefaultSearchRegex = (string) => {
    return new RegExp('.*'+string.toLowerCase()+'.*','g');
};

export const GetColorForAdditionalInfo = (string) => {
    return (string === OpportunityState.NEW
        || string === OpportunityState.WALKTHROUGH_DECLINED
        || string === OpportunityState.NEW_WALKTHROUGH_SCHEDULE_PROPOSED_BY_CLIENT
        || string === OpportunityState.BID_DECLINED
        || string === OpportunityState.NEW_SCHEDULE_PROPOSED_BY_CLIENT)
        ? Color.RED : Color.BORDER_GREY;
};

//Checks if the date is today
export const isToday = (date) => {
    return date.isSame(new Date(), 'day');
};

//Its possible at some point we'll want this to take in params and return options from a start point to an endpoint in given intervals, but for now its a constant
export const GetDurationOptions = () => {
    return {
        60: '1 hour',
        90: '1 hour 30 minutes',
        120: '2 hours',
        150: '2 hour 30 minutes',
        180: '3 hours',
        210: '3 hour 30 minutes',
        240: '4 hours',
        270: '4 hour 30 minutes',
        300: '5 hours',
        330: '5 hour 30 minutes',
        360: '6 hours',
        390: '6 hour 30 minutes',
        420: '7 hours',
        450: '7 hour 30 minutes',
        480: '8 hours',
        510: '8 hour 30 minutes',
        540: '9 hours',
        570: '9 hour 30 minutes',
        600: '10 hours',
        630: '10 hour 30 minutes',
        660: '11 hours',
        690: '11 hour 30 minutes',
        720: '12 hours'
    }
};

// Client's cannot input durations, but they need to be able to see what the vendor had submitted when creating new schedules
// we allow for more duration options in the portal
export const GetClientDurationOptions = () => {
    return {
        15: '15 minutes',
        30: '30 minutes',
        45: '45 minutes',
        60: '1 hour',
        75: '1 hour 15 minutes',
        90: '1 hour 30 minutes',
        105: '1 hour 45 minutes',
        120: '2 hours',
        135: '2 hour 15 minutes',
        150: '2 hour 30 minutes',
        165: '2 hour 45 minutes',
        180: '3 hours',
        195: '3 hour 15 minutes',
        210: '3 hour 30 minutes',
        225: '3 hour 45 minutes',
        240: '4 hours',
        255: '4 hour 15 minutes',
        270: '4 hour 30 minutes',
        285: '4 hour 45 minutes',
        300: '5 hours',
        315: '5 hour 15 minutes',
        330: '5 hour 30 minutes',
        345: '5 hour 45 minutes',
        360: '6 hours',
        375: '6 hour 15 minutes',
        390: '6 hour 30 minutes',
        405: '6 hour 45 minutes',
        420: '7 hours',
        435: '7 hour 15 minutes',
        450: '7 hour 30 minutes',
        465: '7 hour 45 minutes',
        480: '8 hours',
        495: '8 hour 15 minutes',
        510: '8 hour 30 minutes',
        525: '8 hour 45 minutes',
        540: '9 hours',
        555: '9 hour 15 minutes',
        570: '9 hour 30 minutes',
        585: '9 hour 45 minutes',
        600: '10 hours',
        615: '10 hour 15 minutes',
        630: '10 hour 30 minutes',
        645: '10 hour 45 minutes',
        660: '11 hours',
        675: '11 hour 15 minutes',
        690: '11 hour 30 minutes',
        705: '11 hour 45 minutes',
        720: '12 hours',
        735: '12 hour 15 minutes',
        750: '12 hour 30 minutes',
        765: '12 hour 45 minutes',
    }
};

export const GetTimeRangeStrings = (startTime, endTime, minInterval) => {
    let startMoment = startTime instanceof moment ? startTime : moment(startTime, MomentFormat.StandardTime);
    let endMoment = endTime instanceof moment ? endTime : moment(endTime, MomentFormat.StandardTime);

    let momentIntervals = [startMoment];

    while (true) {
        let nextInterval = momentIntervals[momentIntervals.length - 1].clone().add(minInterval, 'minutes');

        if (nextInterval.isSameOrBefore(endMoment))
            momentIntervals.push(nextInterval);
        else
            break;
    }

    return momentIntervals.map((moment) => moment.format(MomentFormat.StandardTime));
};

export const GetRepeatEveryString = (repeatEvery, recurrenceType) => {
    return repeatEvery + ' ' + GetIntervalString(repeatEvery, recurrenceType);
};

export const GetIntervalString = (count, recurrenceType) => {
    switch(recurrenceType) {
        case RecurrenceType.DAILY:
            return count > 1 ? 'days' : 'day';
        case RecurrenceType.WEEKLY:
            return count > 1 ? 'weeks' : 'week';
        case RecurrenceType.MONTHLY:
            return count > 1 ? 'months' : 'month';
    }
};

export const GetRequestRecurringScheduleString = (recurrenceType, repeatEvery, recurrenceCount) => {
    let recurrenceString = recurrenceType;

    //Have to check for this since we have a distinction between RecurrenceType and RfpRecurrenceType
    if (!Object.values(RecurrenceType).includes(recurrenceString))
        recurrenceString = Uppercase(recurrenceString + 'ly');

    switch(recurrenceString) {
        case RecurrenceType.DAILY:
            return 'Repeat every ' + GetRepeatEveryString(repeatEvery, recurrenceString);
        default:
            return 'Repeat every ' + GetRepeatEveryString(repeatEvery, recurrenceString) + ', ' + recurrenceCount + (recurrenceCount > 1 ? ' times' : ' time') + ' per ' + GetIntervalString(1, recurrenceString);
    }
};

export const ConvertDaysOfWeekToListOfNumbers = (sunday, monday, tuesday, wednesday, thursday, friday, saturday) => {
    let daysOfWeekList = [];

    if (sunday)
        daysOfWeekList.push(7);
    if (monday)
        daysOfWeekList.push(1);
    if (tuesday)
        daysOfWeekList.push(2);
    if (wednesday)
        daysOfWeekList.push(3);
    if (thursday)
        daysOfWeekList.push(4);
    if (friday)
        daysOfWeekList.push(5);
    if (saturday)
        daysOfWeekList.push(6);

    return daysOfWeekList;
};

export const ConvertNumberToDayOfWeekShortString = (number) =>{
    switch(number){
        case 1:
            return 'Mon';
        case 2:
            return 'Tue';
        case 3:
            return 'Wed';
        case 4:
            return 'Thu';
        case 5:
            return 'Fri';
        case 6:
            return 'Sat';
        case 7:
            return 'Sun';
        default:
            return 'Invalid Day'
    }
};

export const GetMomentWeekOfMonth = (moment, asString) => {
    let weekOfMonth = Math.ceil(moment.date() / 7);

    if (asString) {
        switch (weekOfMonth) {
            case 1:
                return 'first';
            case 2:
                return 'second';
            case 3:
                return 'third';
            case 4:
                return 'fourth';
        }
    }

    return weekOfMonth;
};

export const GetPathParams = (pathname) => {
    let path = {
        basePath: pathname,
        params: ''
    };

    let secondSlashIndex = path.basePath.indexOf('/', 1);

    if(secondSlashIndex >= 0) {
        path.basePath = pathname.substr(0, secondSlashIndex);
        path.params = pathname.substr(secondSlashIndex).replace('\/','');
    }

    return path;
};

//Removes the number in front of an address and removes any spaces or periods at the end of the address
//This is used for hiding the exact address of a client location.
//example:  906 Conrad St. -> Conrad St
//          24 5th Ave -> 5th Ave
//          Anderson Ln. -> Anderson Ln
export const GetAddressStringOnly = (string) => {
    return string.match(/\d*[a-zA-Z].*[^\s.]/g);
};

//Currently there is a bug in react-native that causes the default margin and padding css styles to not work
//These are workaround functions for ease of use, and will most likely only be used for styled-components since it return a full string of styles
//Top, Right, Bottom, Left
export const Margin = (first, second, third, fourth) => {
    let top = first;
    let bottom = ValueIsSet(third) ? third : first;
    let left = ValueIsSet(fourth) ? fourth : ValueIsSet(second) ? second : first;
    let right = ValueIsSet(second) ? second : first;

    return 'margin-top: ' + top + ';'
        + 'margin-bottom: ' + bottom + ';'
        + 'margin-left: ' + left  + ';'
        + 'margin-right: ' + right + ';'
};

export const Padding = (first, second, third, fourth) => {
    let top = first;
    let bottom = ValueIsSet(third) ? third : first;
    let left = ValueIsSet(fourth) ? fourth : ValueIsSet(second) ? second : first;
    let right = ValueIsSet(second) ? second : first;

    return 'padding-top: ' + top + ';'
        + ' padding-bottom: ' + bottom + ';'
        + ' padding-left: ' + left + ';'
        + ' padding-right: ' + right + ';'
};

export const ConfigureBasicLayoutAnimation = (createDuration, updateDuration, deleteDuration, callback) => {
    LayoutAnimation.configureNext({
        //this duration will apply to all animation types if they aren't explicitly set
        duration: createDuration,

        //Handles animating of newly rendered elements
        //(whatever will be on the page in the next render that isn't there now)
        create: {
            type: LayoutAnimation.Types.linear,
            property: LayoutAnimation.Properties.opacity
        },

        //Handles animating of the elements around the elements that have to be moved/updated
        //(ex. if a block of text appears between two existing blocks of text, the bottom block of text needs to move down to make room)
        update: {
            type: LayoutAnimation.Types.linear,
            duration: ValueIsSet(updateDuration) ? updateDuration : undefined
        },

        //Handles animating of elements that were in the last render, but not in the next one
        //(whatever is on the page now, but won;t be in the next render)
        delete: {
            type: LayoutAnimation.Types.linear,
            property: LayoutAnimation.Properties.opacity,
            duration: ValueIsSet(deleteDuration) ? deleteDuration : undefined
        }
    }, () => callback); //Supposedly this callback is only supported on iOS
};

export const CombineObjects = (...objects) => {
    let combinedObject = {};

    //This for loop structure is the most efficient way of iterating through an array in JavaScript
    for (let i = 0, n = objects.length; i < n; i++) {
        if (objects[i])
            combinedObject = Object.assign(combinedObject, objects[i])
    }

    return combinedObject;
};

//This can be used to debounce functions
//This causes the passed function to not fire off until the interval given has successfully passed without the given function being called
export const Debounce = (fn, interval, ignoreFunction) => {
    let timeout;

    return function() {
        const functionCall = () => fn.apply(this, arguments);

        if (ignoreFunction)
            ignoreFunction();

        clearTimeout(timeout);
        timeout = setTimeout(functionCall, interval);
    }
};

export const GetVersionObject = (versionString) => {
    let versionArr = versionString.split('.');

    return {
        major: versionArr[0],
        minor: versionArr[1],
        patch: versionArr[2],
        build: versionArr.length >= 4 ? versionArr[3] : null
    };
};

export const ObjectsHaveIdenticalValue = (a, b, useLooseEquality) => {
    if (useLooseEquality ? a == b : a === b)
        return true;

    let typeOfA = typeof a;
    let typeOfB = typeof b;

    //If a and b aren't the same type then we will say they can't possibly be equivalent in value
    if (typeOfA !== typeOfB)
        return false;

    switch(typeOfA) {
        //If the very first a === b statement didn't return then the boolean values are definitely not the same
        case JavascriptType.BOOLEAN:
            return false;

        //For now if a === b didn't pass then we will say that the functions are different. We can build this out later if its needed
        case JavascriptType.FUNCTION:
            return a.toString() === b.toString();

        //If a === b didn't pass then the numbers are different
        case JavascriptType.NUMBER:
            return false;

        //For objects we need to go through all the keys of a and b and make sure they are identical
        case JavascriptType.OBJECT:
            let aKeys = Object.keys(a);
            let bKeys = Object.keys(b);

            if (aKeys.length !== bKeys.length)
                return false;
            else if (aKeys.length === 0)
                return true;
            else {
                aKeys.sort();
                bKeys.sort();

                return aKeys.every((key, index) => {
                    if (key !== bKeys[index])
                        return false;

                    return ObjectsHaveIdenticalValue(a[key], b[key], useLooseEquality);
                })
            }

        //If a === b didn't pass then the strings are different
        case JavascriptType.STRING:
            return false;

        //This should cover symbol comparison
        case JavascriptType.SYMBOL:
            return a.toString() === b.toString();

        //If we got here that means both a and b are `undefined` so they are equal
        case JavascriptType.UNDEFINED:
            return true;
    }

    //This would only happen if the type of both a and b wasn't a Javascript primitive, so we return false since we can't guess at how to evaluate equivalency
    return false;
};

//Returns the first index if any value in the given array is identical in value to the given value param, otherwise it returns -1
export const ArrayIndexOfEquivalentValue = (array, value) => {
    let index = -1;

    array.some((item, i) => {
        if (ObjectsHaveIdenticalValue(item, value)) {
            index = i;
            return true;
        }

        return false;
    });

    return index;
};

export const canOpenUrl = (url, openIfTrue) => {
    return (
        Linking.canOpenURL(url).then(result => {
            if (!!result && openIfTrue) {
               Linking.openURL(url);
               return true;
            } else {
               return (!!result);
            }
        }).catch(() => {
            return false;
        })
    )
};

export const GetDeviceSafeViewOffset = () => {
    let deviceModel = DeviceInfo.getModel();

    if (deviceModel.includes('iPhone X'))
        return 34;

    return undefined;
};
